/*
  Nanoameter code range down to 1nA and +/- 670 uA with a 100 ohm shunt.
  the shunt value can be changed if need to measure bigger amperages.

  +/- 60mA with a 1 ohm shunt.Protect with a 500mA fuse

  This is designed as test code and is ugly bug.

  Many immprovements can be made - structure and serial comms - but it works
  code is repeated and can be cleaned

  Serial comms 115200

  Autorange for 60mA or low down to 1nA

 ****YOU MUST STORE CAL VALUE IN EEPROM ON FIRST RUN - press d for defaults***

  ----------------------------------------
  Pins used
  Pin 2 used to control range relay
  LiquidCrystal lcd(8, 9, 4, 5, 6, 7) and A0 for keys

  ----------------------------------------
  Buttons:

  Down button =stop button
  Up button = display cal display values button
  Left button = run button
  Right button = default cal values button
  Select button = zero/offset button
  ---------------------------------------

  To calibrate use c and h commands.
  c is low range use test current of 200uA
  q is high range use test current of 1mA

  With calibration can be better than 0.1%

*/

#include "HX711.h" // need this library
#include <EEPROM.h>

#include <LiquidCrystal.h>

// initialize the library with the numbers of the interface pins
LiquidCrystal lcd(8, 9, 4, 5, 6, 7);

#define calEEPROM 0
#define calHighEEPROM 10
#define offNa 0;
#define REF 1252 // 1252 determined by experiment and is main calibration - change here or can change during run time.
#define HighCal 1.15
#define default_offset_low 20563.63 //default dc offsetVlaue found by testing a few units but will be automatically 
#define default_offset_high 150 //default dc offsetVlaue found by testing a few units but will be automatically 

//  determined on start up.

// used pins associated with the HX711
// HX711.DOUT	- pin #A2
// HX711.PD_SCK	- pin #A1

HX711 scale;		// parameter "gain" is ommited; the default value 128 is used by the library

// general registers used in claclulation
uint32_t tmp_dut; // raw value
char incomingByte; //serial port
uint32_t count = 1; // sample counter
float ampstemp; //temp calculated result
const int relay = 2; // relay control pin


// flags used to define behaviour
boolean ampsFlag = false; //run flag
boolean engFlag = false; //engineering data dump raw values flag
boolean relayOnFlag = false;

// calibration registers
float calValue = REF; // nominal scaling factor
float offsetValueLow = default_offset_low; //dc mid point
float offsetValueHigh = default_offset_high; //dc mid point
float offSetNa = 0; //nano amp offset after calulating dc mid point (typically 20nA)
float highCalReg = HighCal;
float offsetValue;

// flags to control the behaviour via buttons
boolean  select_button = false;
boolean  left_button = false;
boolean  right_button = false;
boolean  up_button = false;
boolean  down_button = false;

// prototype functions
void serial_comms(); // serial comms
void messages();
boolean Analogue_Button(void);
void test_keys(void);

// not used but here just in case ----
void EEPROMWrite16bit(int address, uint16_t value);
uint16_t EEPROMRead16bit(int address);
void EEPROMWritelong(int address, long value);
long  EEPROMReadlong(int address);
// -----------------------------------
void writeFloat(unsigned int addr, float x);
float readFloat(unsigned int addr);

void reset_CPU();  // CPU hard reset if all goes wrong - could take this out

void Zero(); // Zero out DC point and nanoamp offest

// Start of code
void setup() {
  pinMode(relay, OUTPUT);
  digitalWrite(relay, LOW);
  Serial.begin(115200); // set serial
  // initialize the serial communications:
  // set up the LCD's number of columns and rows:
  lcd.begin(16, 2);
  // Print a message to the LCD.
  lcd.setCursor(0, 0);
  lcd.print("Nano Ampmeter");
  lcd.setCursor(0, 1);
  lcd.print("Preparing");
  Serial.println(F("Nano Ampmeter"));
  Serial.print(F("Code build date ")); //Indicate date built
  Serial.print(__TIME__);
  Serial.print(F("  "));
  Serial.println (__DATE__);
  Serial.println(F(" "));
  Serial.println(F(" Type Q for list of commands"));
  Serial.println(F(" ------------------------------"));
  delay(4000); //splash screen

  calValue  = readFloat(calEEPROM); // read cal value back from EEPROM
  highCalReg = readFloat(calHighEEPROM);
  scale.begin(A2, A1, 32); //set ports and gains
  tmp_dut = scale.read();      // get it started
  messages(); //displays menu
  Zero(); // initialise the system and offset cancellation

  // print ready
  Serial.println(F("Measure Current "));
  Serial.println(F(" "));
}

void loop() {

  boolean negative; // used to indicate negative value
  test_keys();
  if (relayOnFlag)
  {
    digitalWrite(relay, HIGH);
    offsetValue = offsetValueHigh;
  }
  else {
    digitalWrite(relay, LOW);
    offsetValue = offsetValueLow;
  }

  // serial stuff to read keyboard could be improved
  if (Serial.available() > 0) {
    // read the incoming byte:
    incomingByte = Serial.read();
    serial_comms();
  }

  // Calculation bit of the software

  // if in measurement mode
  if (ampsFlag) {
    //read channel B
    Serial.print(count);
    Serial.print(F(","));

    // read raw value form DUT
    scale.set_gain(32); //Set next conversion to be ch B, 32x gain
    tmp_dut = scale.read_average(32); //read ref on CH B, and set next conversion to be ch B, 32x gain

    // check if negative and if so two's complement conversion
    if (tmp_dut > 4000000000) {
      tmp_dut  = (~tmp_dut ) + 1; // two's complement conversion if negative value
      negative = true; // indicate positive for no minus sign
    }
    else negative = false; //indicate negative for minus sign

    // calulate meaured value
    ampstemp = (tmp_dut - offsetValue);
    if (!negative) ampstemp = (tmp_dut - offsetValue);
    else ampstemp = (tmp_dut + offsetValue);

    // if in engineering mode dump raw values
    if (engFlag)
    {
      Serial.print(F("Raw Read = "));
      Serial.println(tmp_dut);
      Serial.print(F("Offset processed = "));
      Serial.println(ampstemp, 6);
    }

    // nano amp range if less than 1uA
    if (ampstemp < (calValue * 10) && !relayOnFlag)
    {
      ampstemp  = ((ampstemp * 100) / calValue) - offSetNa;
      if (negative) Serial.print(F("-"));
      Serial.print(ampstemp  , 5);
      Serial.println(F(" nA"));
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print(F("Measuring"));
      lcd.setCursor(0, 1);

      if (negative) lcd.print(F("-"));
      lcd.print(ampstemp, 5);
      lcd.print(F(" nA"));
    }
    else // microamp range
    {
      ampstemp  = ((ampstemp) / (calValue * 10)) - (offSetNa / 1000);

      if (negative) Serial.print(F("-"));

      if (abs(ampstemp) > 600) relayOnFlag = true;
      if (abs(ampstemp) < 3) relayOnFlag = false;

      if (relayOnFlag)
      {
        ampstemp = ampstemp * highCalReg / 10;
        Serial.print(ampstemp, 5);
        Serial.println(F(" mA"));
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(F("Measuring"));
        lcd.setCursor(0, 1);
        if (negative) lcd.print(F("-"));
        lcd.print(ampstemp, 5);
        lcd.print(F(" mA"));
      }
      else
      {
        Serial.print(ampstemp, 5);
        Serial.println(F(" uA"));
        lcd.clear();
        lcd.setCursor(0, 0);
        lcd.print(F("Measuring"));
        lcd.setCursor(0, 1);
        if (negative) lcd.print(F("-"));
        lcd.print(ampstemp, 5);
        lcd.print(F(" uA"));
      }
    }
    count++; // used to out put count for CSV for excel
  }
}
// Communication routine used to respond to remote commands via the serial port
// clunky but works
void serial_comms()
{
  String value;
  float localAv = 0;


  char temp = tolower(incomingByte);// remove case sensitivity
  switch (temp) // test the value of the serial character
  {
    case 'c':// calibrate reference resistor (10ohms)
      Serial.println(F(" "));
      Serial.println(F("To calibrate enter value of scaling factor"));
      Serial.println(F("e.g. 1252 as 1252"));
      ampsFlag = false;
      engFlag = false;
      while (Serial.available() == 0) {           //Wait for user input
      }
      value = Serial.readString();
      calValue  = value.toFloat();
      Serial.print(F("Calibration factor low range= "));
      Serial.println(calValue);
      writeFloat(calEEPROM, calValue);
      break;

    case 'h':// calibrate high range reference resistor (1ohms)
      Serial.println(F(" "));
      Serial.println(F("To calibrate high value of scaling factor"));
      Serial.println(F("e.g. 1.15 as 1150"));
      ampsFlag = false;
      engFlag = false;
      while (Serial.available() == 0) {           //Wait for user input
      }
      value = Serial.readString();
      highCalReg  = value.toFloat();
      highCalReg /= 1000;
      Serial.print(F("Calibration factor high range= "));
      Serial.println(highCalReg, 4);
      writeFloat(calHighEEPROM, highCalReg); // write high cal value
      break;


    case 's':// Stop and clear all falgs etc
      Serial.println(F(" "));
      Serial.println(F("Stop"));
      Serial.println(F("Reset Flags to Defaults"));
      ampsFlag = false;
      engFlag = false;
      relayOnFlag = false;
      digitalWrite(relay, LOW);
      count = 1; // reset count value
      break;

    case 'r':// Run with last settings for offset
      Serial.println(F(" "));
      Serial.println(F("Run"));
      Serial.println(F("Read Current"));
      ampsFlag = true;
      engFlag = false;
      count = 1; // reset count value
      break;

    case 'e':// Engineers mode  - output raw data
      Serial.println(F(" "));
      Serial.println(F("Engineer"));
      Serial.println(F("Output all raw data"));
      engFlag = true;
      break;

    case 'd':// default cal values - should run this the first time so defauls are correct in memory at restart
      Serial.println(F(" "));
      Serial.println(F("Offset and Calibration set to default"));
      Serial.println(F("Default cal"));

      calValue = REF;
      highCalReg = HighCal;
      offsetValueLow = default_offset_low;
      offsetValueHigh = default_offset_high;
      offSetNa = offNa; // use current value
      engFlag = false;
      writeFloat(calEEPROM, calValue);
      writeFloat(calHighEEPROM, highCalReg);
      Serial.print(F("Current Cal Value = "));
      Serial.println(calValue, 4);
      Serial.print(F("Current Cal Value High = "));
      Serial.println(highCalReg, 4);
      Serial.print(F("Current Offset Value = "));
      Serial.println(offsetValueLow);
      Serial.print(F("Current Offset Value = "));
      Serial.println(offsetValueHigh);
      break;

    case '*' :// complete reset!! restore all default conditions
      Serial.println(F("Complete reset!! restore all default conditions "));
      Serial.println(F(" "));
      Serial.println(F("Reset Now"));
      delay(2000);
      reset_CPU(); // reset the hardware of the systems
      break;

    case 'i' :// Calculta the offsets and DC point
      Zero();
      break;

    case 'q' : //help info display menu items
      messages();
      break;
  }
}

// EEPROM stuff
//This function will write a 4 byte (32bit) long to the eeprom at
//the specified address to address + 3.
void EEPROMWritelong(int address, long value)
{
  //Decomposition from a long to 4 bytes by using bitshift.
  //One = Most significant -> Four = Least significant byte
  byte four = (value & 0xFF);
  byte three = ((value >> 8) & 0xFF);
  byte two = ((value >> 16) & 0xFF);
  byte one = ((value >> 24) & 0xFF);

  //Write the 4 bytes into the eeprom memory.
  EEPROM.write(address, four);
  EEPROM.write(address + 1, three);
  EEPROM.write(address + 2, two);
  EEPROM.write(address + 3, one);
}

//Recall Eeprom
long EEPROMReadlong(int address)
{
  //Read the 4 bytes from the eeprom memory.
  long four = EEPROM.read(address);
  long three = EEPROM.read(address + 1);
  long two = EEPROM.read(address + 2);
  long one = EEPROM.read(address + 3);

  //Return the recomposed long by using bitshift.
  return ((four << 0) & 0xFF) + ((three << 8) & 0xFFFF) + ((two << 16) & 0xFFFFFF) + ((one << 24) & 0xFFFFFFFF);
}

//Recall Eeprom
uint16_t EEPROMRead16bit(int address)
{
  //Read the 4 bytes from the eeprom memory.
  long four = EEPROM.read(address);
  long three = EEPROM.read(address + 1);
  //Return the recomposed long by using bitshift.
  return (((four << 0) & 0xFF) + ((three << 8) & 0xFFFF));
}

// These two routines are used
void EEPROMWrite16bit(int address, uint16_t value)
{
  //Decomposition from a long to 4 bytes by using bitshift.
  //One = Most significant -> Four = Least significant byte
  byte four = (value & 0xFF);
  byte three = ((value >> 8) & 0xFF);


  //Write the 4 bytes into the eeprom memory.
  EEPROM.write(address, four);
  EEPROM.write(address + 1, three);
}

float readFloat(unsigned int addr)
{
  union
  {
    byte b[4];
    float f;
  } data;
  for (int i = 0; i < 4; i++)
  {
    data.b[i] = EEPROM.read(addr + i);
  }
  return data.f;
}
void writeFloat(unsigned int addr, float x)
{
  union
  {
    byte b[4];
    float f;
  } data;
  data.f = x;
  for (int i = 0; i < 4; i++)
  {
    EEPROM.write(addr + i, data.b[i]);
  }
}

// Cause CPU to reset
void reset_CPU() // Restarts program from beginning but does not reset the peripherals and registers
{
  asm volatile ("  jmp 0");
}

// display commands
void messages()
{
  Serial.println(F("Firmware uses the HX711 - version 1.0"));
  Serial.println(F("Measure nano amps"));

  Serial.println(F(""));
  Serial.println(F("Menu - help = Q"));
  Serial.println(F(""));
  Serial.println(F("Stop  = s"));
  Serial.println(F("Stop current process"));
  Serial.println(F(""));
  Serial.println(F("Run = r"));
  Serial.println(F("Read Current"));
  Serial.println(F(""));
  Serial.println(F("Cal  = c"));
  Serial.println(F("Calibrate scale factor"));
  Serial.println(F(""));
  Serial.println(F("Cal  = h"));
  Serial.println(F("Calibrate scale factor high range"));
  Serial.println(F(""));
  Serial.println(F("Default Cal = d"));
  Serial.println(F("Set to default values"));
  Serial.println(F(""));
  Serial.println(F("Reset Micro=  *"));
  Serial.println(F("Reset all and reboot"));
  Serial.println(F(""));
  Serial.println(F("Measure Current  =  i "));
  Serial.println(F("Measure Nano/Microamps"));
  Serial.println(F(""));
  Serial.println(F("Engineer's Mode =  e "));
  Serial.println(F("Raw Values"));
  Serial.println(F(""));
  Serial.println(F("Calibrate Values"));
  Serial.print(F("Cal Value = "));
  Serial.println(calValue, 4);
  Serial.print(F("Cal Value = "));
  Serial.println(highCalReg, 4);
  Serial.print(F("Current Offset Value = "));
  Serial.println(offSetNa, 6);
}


// Calculate DC point and nano amp offset.
// best done with the leads in place and left open circuit

void Zero()
{
  float localAv = 0;
  // clear control flags
  count = 0;
  ampsFlag = true;
  engFlag = false;

  //start measuring
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(F("Initialising"));
  Serial.println(F("Initialising Offsets and DC point"));
  digitalWrite(relay, LOW);
  delay(1000);
  Serial.println(F("Measure low level"));
  Serial.println(F(" "));
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(F("Low DC Offset"));

  localAv = 0;
  lcd.setCursor(0, 1);
  //Calibrate low
  for ( int i = 0; i < 16; i++)
  {
    scale.set_gain(32); //read ref on CH A, 64x gain. Set next conversion to be ch B, 32x gain
    tmp_dut = scale.read_average(8); //read ref on CH A, 64x gain. Set next conversion to be ch B, 32x gain

    if (tmp_dut > 4000000000) {   // negative
      tmp_dut  = (~tmp_dut ) + 1; // two's complement conversion if negative value
      localAv -= tmp_dut;         // need this if it happens to go negative otherwise get wrong value
    }
    else localAv += tmp_dut;

    Serial.print(F(".")); //show something is happening
    lcd.print(F("."));
  }
  offsetValueLow = localAv / 16;

  Serial.println(F(" "));
  Serial.print(F("Zero Value Low = "));
  Serial.println(offsetValueLow);
  Serial.println(F(" "));
  lcd.setCursor(0, 1);
  lcd.print(offsetValueLow);

  delay(1000);

// nA offset calculation
  scale.set_gain(32); //read ref on CH B, 32x gain. Set next conversion to be ch B, 32x gain
  tmp_dut = scale.read_average(8); //read ref on CH B, 32x gain. Set next conversion to be ch B, 32x gain
  ampstemp = (tmp_dut - offsetValueLow);

  ampstemp  = (ampstemp * 100) / calValue;
  offSetNa = ampstemp;

  Serial.print(ampstemp , 4);
  Serial.println(F(" offset low nA"));
  Serial.println(F(" "));
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(F("Low nA Offset"));
  lcd.setCursor(0, 1);
  lcd.print(offSetNa, 3);
  lcd.print(F(" nA offset "));

  // High range calculation
  // set relay on
  digitalWrite(relay, HIGH);
  delay(2000);

  Serial.println(F("Measure high level"));
  Serial.println(F(" "));
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(F("High DC Offset"));

  localAv = 0;
  lcd.setCursor(0, 1);
  //Calibrate High
  for ( int i = 0; i < 16; i++)
  {
    scale.set_gain(32); //read ref on CH B, 32x gain. Set next conversion to be ch B, 32x gain
    tmp_dut = scale.read_average(8); //read ref on CH B, 32x gain. Set next conversion to be ch B, 32x gain

    if (tmp_dut > 4000000000) {   // negative
      tmp_dut  = (~tmp_dut ) + 1; // two's complement conversion if negative value
      localAv -= tmp_dut;         // need this if it happens to go negative otherwise get wrong value
    }
    else localAv += tmp_dut;
    
    Serial.print(F(".")); //show something is happening
    lcd.print(F("."));
  }
  offsetValueHigh = localAv / 16;
  Serial.println(F(" "));
  Serial.print(F("Zero Value High = "));
  Serial.println(offsetValueHigh);
  lcd.setCursor(0, 1);
  lcd.print(offsetValueHigh);
  Serial.println(F("  Ready to Measure Current"));
  Serial.println(F(" "));
  digitalWrite(relay, LOW);
  delay(2000);
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print(F("Measure Current"));
}

// Test keys and respond to press
void test_keys(void) {
  if (Analogue_Button()) {
    if (up_button == true)// display cal values button
    {
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print(F("LowCal= "));
      lcd.print(calValue, 3);
      lcd.setCursor(0, 1);
      lcd.print(F("HighCal= "));
      lcd.print(highCalReg, 3);
      delay(1000);
    }
    else if (down_button == true)// stop button
    {
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print(F("Stop"));
      delay(1000);
      Serial.println(F(" "));
      Serial.println(F("Stop"));
      Serial.println(F("Reset Flags to Defaults"));
      ampsFlag = false;
      engFlag = false;
      relayOnFlag = false;
      digitalWrite(relay, LOW);
      count = 1; // reset count value
    }
    else if (left_button == true)// run button
    {
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print(F("Run"));
      delay(1000);
      Serial.println(F(" "));
      Serial.println(F("Run"));
      Serial.println(F("Read Current"));
      ampsFlag = true;
      engFlag = false;
      count = 1; // reset count value
    }
    else if (right_button == true)// default cal button
    {
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print(F("Default cal"));
      delay(1000);
      Serial.println(F(" "));
      Serial.println(F("Offset and Calibration set to default"));
      Serial.println(F("Default cal"));
      calValue = REF;
      offsetValueLow = default_offset_low;
      offsetValueHigh = default_offset_high;
      offSetNa = offNa; // use current value
      highCalReg = HighCal;
      engFlag = false;
      delay(2000);
      writeFloat(calEEPROM, calValue);
      writeFloat(calHighEEPROM, highCalReg);
      Serial.print(F("Current Cal Value = "));
      Serial.println(calValue, 4);
      Serial.print(F("Current Cal Value High= "));
      Serial.println(highCalReg, 4);
      Serial.print(F("Current Offset Value Low = "));
      Serial.println(offsetValueLow);
      Serial.print(F("Current Offset Value High = "));
      Serial.println(offsetValueHigh);
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print(F("LowCal= "));
      lcd.print(calValue);
      lcd.setCursor(0, 1);
      lcd.print(F("HighCal= "));
      lcd.print(highCalReg);
      delay(1000);
    }
    else if (select_button == true)// zero button
    {
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print(F("Zero/Offset"));
      delay(1000);
      Zero();
    }
  }
}

// Read Analogue buttons via A/D flags used to indicate the state
boolean Analogue_Button() {

  // keys
  select_button = false;
  left_button = false;
  right_button = false;
  up_button = false;
  down_button = false;
  int x;
  x = analogRead (0);

  if (x < 30) { // right
    right_button = true;
  }
  else if (x > 30 && x < 150) { // up
    up_button = true;
  }
  else if (x > 150 && x < 350) { //down
    down_button = true;
  }
  else if (x > 420 && x < 600) { // left
    left_button = true;
  }
  else if (x > 600 && x < 800) { // select
    select_button = true;
  }
  if (select_button == true || left_button == true || right_button == true || down_button || up_button) return (true);
  else return (false);
}



